use super::board::Pos;

pub type MarbleValue = u32;

#[derive(Debug, Default, Clone, Copy, PartialEq)]
pub enum Tile {
	#[default]
	Blank,
	Block,
	Marble {
		value: MarbleValue,
		dir: Direction,
	},
	Digit(u8),
	Mirror(MirrorType),
	Arrow(Direction),
	Powerable(PTile, bool),
}

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum PTile {
	Trigger,
	Wire(WireType),
	Gate(GateType),
	Math(MathOp),
	Bag,
	Flipper,
	Input,
	Output,
}

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum MirrorType {
	Forward,
	Back,
}

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum MathOp {
	Add,
	Sub,
	Mul,
	Div,
	Rem,
}

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum GateType {
	LessThan,
	GreaterThan,
	Equal,
	NotEqual,
}

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum WireType {
	Vertical,
	Horizontal,
	Cross,
}

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum Direction {
	Up,
	Down,
	Left,
	Right,
}

impl Tile {
	pub fn is_blank(&self) -> bool {
		matches!(self, Tile::Blank)
	}

	pub fn read_value(&self) -> MarbleValue {
		if let Tile::Marble { value, dir: _ } = self {
			*value
		} else {
			0
		}
	}

	pub fn texture(&self) -> String {
		match self {
			Tile::Blank => "transparent",
			Tile::Block => "block",
			Tile::Marble { value: _, dir: _ } => "marble",
			Tile::Digit(n) => match n {
				b'0' => "digit_0",
				b'1' => "digit_1",
				b'2' => "digit_2",
				b'3' => "digit_3",
				b'4' => "digit_4",
				b'5' => "digit_5",
				b'6' => "digit_6",
				b'7' => "digit_7",
				b'8' => "digit_8",
				b'9' => "digit_9",
				_ => unreachable!("invalid digit"),
			},
			Tile::Mirror(mirror) => match mirror {
				MirrorType::Forward => "mirror_forward",
				MirrorType::Back => "mirror_back",
			},
			Tile::Arrow(dir) => match dir {
				Direction::Up => "arrow_up",
				Direction::Down => "arrow_down",
				Direction::Left => "arrow_left",
				Direction::Right => "arrow_right",
			},
			Tile::Powerable(tile, state) => {
				let root = match tile {
					PTile::Trigger => "trigger",
					PTile::Wire(wire) => match wire {
						WireType::Vertical => "wire_vertical",
						WireType::Horizontal => "wire_horizontal",
						WireType::Cross => "wire_cross",
					},
					PTile::Gate(gate) => match gate {
						GateType::LessThan => "lt",
						GateType::GreaterThan => "gt",
						GateType::Equal => "eq",
						GateType::NotEqual => "neq",
					},
					PTile::Math(math_op) => match math_op {
						MathOp::Add => "add",
						MathOp::Sub => "sub",
						MathOp::Mul => "mul",
						MathOp::Div => "div",
						MathOp::Rem => "rem",
					},
					PTile::Bag => "bag",
					PTile::Flipper => "flipper",
					PTile::Input => "input",
					PTile::Output => "output",
				};
				return format!("{root}_{}", if *state { "on" } else { "off" });
			}
		}
		.to_owned()
	}
}

impl Direction {
	pub const ALL: [Direction; 4] = [
		Direction::Up,
		Direction::Down,
		Direction::Left,
		Direction::Right,
	];

	pub fn opposite(&self) -> Direction {
		match self {
			Direction::Up => Direction::Down,
			Direction::Down => Direction::Up,
			Direction::Left => Direction::Right,
			Direction::Right => Direction::Left,
		}
	}

	pub fn right(&self) -> Direction {
		match self {
			Direction::Up => Direction::Right,
			Direction::Down => Direction::Left,
			Direction::Left => Direction::Up,
			Direction::Right => Direction::Down,
		}
	}

	pub fn left(&self) -> Direction {
		self.right().opposite()
	}

	pub fn step(&self, mut pos: Pos) -> Pos {
		match self {
			Direction::Up => pos.y -= 1,
			Direction::Down => pos.y += 1,
			Direction::Left => pos.x -= 1,
			Direction::Right => pos.x += 1,
		}
		pos
	}
}

impl WireType {
	pub fn directions(self) -> &'static [Direction] {
		match self {
			WireType::Vertical => &[Direction::Up, Direction::Down],
			WireType::Horizontal => &[Direction::Left, Direction::Right],
			WireType::Cross => &Direction::ALL,
		}
	}
}

impl MirrorType {
	pub fn new_dir(self, dir: Direction) -> Direction {
		match self {
			MirrorType::Forward => match dir {
				Direction::Up => Direction::Right,
				Direction::Down => Direction::Left,
				Direction::Left => Direction::Down,
				Direction::Right => Direction::Up,
			},
			MirrorType::Back => match dir {
				Direction::Up => Direction::Left,
				Direction::Down => Direction::Right,
				Direction::Left => Direction::Up,
				Direction::Right => Direction::Down,
			},
		}
	}
}
